package org.systemsbiology.jrap;

import java.util.zip.DataFormatException;
import java.util.zip.Inflater;

public class Bytes {
	
	/*
	 Jimmy Eng 
	 The values are stored as network byte order (big endian) before base64
	 encoding.
	 http://groups.google.com/group/spctools-discuss/browse_thread/thread/eb6b46d853946f32
	 */
		 
	// http://www.devx.com/tips/Tip/34353
	// 2-byte number
	public static int SHORT_little_endian_TO_big_endian(int i)
	{
	    return ((i>>8)&0xff)+((i << 8)&0xff00);
	}

	// http://www.devx.com/tips/Tip/34353
	// 4-byte number
	public static int INT_little_endian_TO_big_endian(int i)
	{
	    return((i&0xff)<<24)+((i&0xff00)<<8)+((i&0xff0000)>>8)+((i>>24)&0xff);
	}
	
	public static float quadletToFloat(byte... bytes) {
		
		// byte1 byte2 byte3 byte4 -> integer (4 bytes) -> float
		int intBits = 0;
		for (int i=0 ; i<4 ; i++) {
			intBits |= (((int) bytes[i]) & 0xff);
			if (i<3) {
				intBits <<= 8;
			}
		}
		return Float.intBitsToFloat(intBits);
	}
	
	public static double octletToDouble(byte... bytes) {
		
		// byte1 byte2 byte3 byte4 byte5 byte6 byte7 byte8-> long (8 bytes) -> double
		long longBits = 0L;
		for (int i=0 ; i<8 ; i++) {
			longBits |= (((long) bytes[i]) & 0xff);
			if (i<7) {
				longBits <<= 8;
			}
		}
		return Double.longBitsToDouble(longBits);
	}
	
	public static byte[] floatToQuadlet(float real) {
		
		int intBits = Float.floatToIntBits(real);
		
		byte[]ba = new byte[4];
		
		for (int i=ba.length-1 ; i>=0 ; i--) {
			ba[i] = (byte)(intBits & 0xff);		
			intBits >>= 8;
		}
		
		return ba;
	}
	
	public static byte[] doubleToOctlet(double real) {
		
		long longBits = Double.doubleToLongBits(real);
		
		byte[]ba = new byte[8];
		
		for (int i=ba.length-1 ; i>=0 ; i--) {
			ba[i] = (byte)(longBits & 0xff);		
			longBits >>= 8;
		}
		
		return ba;
	}
	
	/**
	 * This method convert chunks of bytes into couple of floats.
	 * 
	 * @param bytes byte-codes spectrum
	 * @param precision mz and intensity precision in bytes
	 * @return
	 */
	public static float[][] byteQuadsToMzInts(byte[] bytes, int peakNumber, int precision) {
		// number of couple mz-int ()
		int mzIntNumber = bytes.length / (precision * 2);
		
		if (mzIntNumber != peakNumber) {
			throw new IllegalArgumentException("bad number of peaks: expected "
					+ peakNumber + ", found "+mzIntNumber);
		}
		
		float[][] tmpMassIntensityList = new float[2][mzIntNumber];
		int peakIndex = 0;
		int fieldIndex = 0;

		if (precision <= 0)
			System.err.println("FLOATBYTES <= 0!!!");

		/* TODO: do not understand why they do not read the last float ???? */
		//for (int i = 0; i < bytes.length - floatBytes; i += precision)
		
		for (int i=0 ; i<bytes.length ; i += precision)
		{
			float result = 0;
			
			if (precision == 4) {
				result = quadletToFloat(bytes[i], bytes[i+1], bytes[i+2], bytes[i+3]);
			} else {
				result = (float) octletToDouble(bytes[i], bytes[i+1], bytes[i+2], bytes[i+3],
						bytes[i+4], bytes[i+5], bytes[i+6], bytes[i+7]);
			}
			
			// BUG FIXED: for a precision higher than 4 bytes all the other bytes will be ignored
			// Question: is in the double(8 bytes)'s 4 first bytes part lives float ?
			// It seems to be the case in the standard IEEE 754
			// http://fr.wikipedia.org/wiki/IEEE_754
			//
			// Must be in IEEE 754 encoding!
			// QUID of big endian vs little endian data management ?
			tmpMassIntensityList[fieldIndex++][peakIndex] = result;
			
			if (fieldIndex == 2)
			{
				fieldIndex = 0;
				peakIndex++;
			}
		}
		
		bytes = null;
		
		return tmpMassIntensityList;
	}

	// zlib uncompression
	// from an example at http://java.sun.com/developer/technicalArticles/Programming/compression/
	// see ramp.c
	public static byte[] uncompress(byte[] bytes, int compressedLen, int uncompressedLen) throws DataFormatException {
		
      //  pUncompr = (Byte*)calloc((uInt) uncompressedLen , 1);
		
      //  err = uncompress( pUncompr , &uncompressedLen , (const Bytef*)pDecoded , compressedLen );

		// Decompress the bytes
		Inflater decompresser = new Inflater();
		decompresser.setInput(bytes, 0, compressedLen);
		byte[] result = new byte[uncompressedLen];
		//int resultLength = 
		decompresser.inflate(result);
		decompresser.end();
		
		//System.out.println("inflate size="+resultLength);

		return result;
	}
}
